package net.sf.buildbox.util;

/**
 * Helper class for producers that are periodically scanning some input
 */
public abstract class AbstractSampler<T> {

    /**
     * Check if there is an object available and pull it.
     * @return pulled object or null if none available.
     */
    protected abstract T scan();

    /**
     * Wait some time for an object
     * @param timeout  total time limit; 0 means try just once, less than 0 means no time limit
     * @param quantum  scanning period, should be much less than timeout
     * @return pulled object or null if none was available all the time.
     */
    public T produceObject(long timeout, long quantum) {
        long time0 = System.currentTimeMillis();
        long endTime = time0 + timeout;
        T object = scan();
        long scanTime = System.currentTimeMillis() - time0;
        if (timeout == 0) return object; // try once
        while (object == null) {
            long sleeptime;
            if (timeout < 0) {
                // infinite waiting - we don't check endTime here
                sleeptime = quantum;
            } else {
                sleeptime = endTime - System.currentTimeMillis();
                if (sleeptime > quantum) {
                    sleeptime = quantum;
                } else if (sleeptime <= 0) {
                    break;
                }
            }
            if (sleeptime > scanTime) {
                try {
                    Thread.sleep(sleeptime - scanTime);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                    return null;
                }
            }
            if (Thread.interrupted()) {
                Thread.currentThread().interrupt();
                return null;
            }
            time0 = System.currentTimeMillis();
            object = scan();
            scanTime = System.currentTimeMillis() - time0;
        }
        return object;
    }
}
