package net.sf.buildbox.util;

import java.util.Map;
import java.util.HashMap;
import java.util.Enumeration;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Constructor;

/**
 * Purpose: to cleanly isolate platform specific code.
 */
public class BbxPlatformUtils {
    private static Platform currentPlatform;
    public static Platform getCurrentPlatform() {
        if (currentPlatform == null) {
            // try to detect
            final String osName = System.getProperty("os.name").toUpperCase();
            if (osName.startsWith("WIN")) {
                currentPlatform = Platform.WINDOWS;
            } else if (osName.startsWith("LIN")) {
                currentPlatform = Platform.LINUX;
            } else if (osName.startsWith("MAC")) {
                currentPlatform = Platform.MAC;
            } else {
                throw new UnsupportedOperationException("Unknown platform: os.name=" + osName);
            }
        }
        return currentPlatform;
    }

    public static class PlatformSwitch<PH> {
        private Map<Platform, Class<? extends PH>> handlerClasses = new HashMap<Platform, Class<? extends PH>>();
        private Map<Platform, PH> handlerCache = new HashMap<Platform, PH>();

        public final PH getCurrentPlatformHandler() {
            return getPlatfromHandler(getCurrentPlatform());
        }

        public PH getPlatfromHandler(Platform platform) {
            PH handler = handlerCache.get(platform);
            if (handler == null) {
                final Class<? extends PH> handlerClass = handlerClasses.get(platform);
                if (handlerClass == null) {
                    // TODO: support some sort of fallback support, like LINUX -> UNIX etc.
                    throw new UnsupportedOperationException("Unsupported platform: " + platform);
                }
                try {
                    final Constructor<? extends PH> constructor = handlerClass.getDeclaredConstructor();
                    constructor.setAccessible(true);
                    handler = constructor.newInstance();
                    handlerCache.put(platform, handler);
                } catch (InstantiationException e) {
                    throw new IllegalStateException(e);
                } catch (IllegalAccessException e) {
                    throw new IllegalStateException(e);
                } catch (InvocationTargetException e) {
                    throw new IllegalStateException(e);
                } catch (NoSuchMethodException e) {
                    throw new IllegalStateException(e);
                }
            }
            return handler;
        }

        public void setPlatformHandler(Platform platform, Class<? extends PH> handlerClass) {
            handlerClasses.put(platform, handlerClass);
        }
    }

    public static enum Platform {
        LINUX, WINDOWS, MAC, UNIX
    }

    /**
     * Displays system properties, system environment and other platform-related stuff so that it can be used to gather specific platform behaviour.
     * Usage:
     * java -cp buildbox-util.jar {@link net.sf.buildbox.util.BbxPlatformUtils}
     * @param args no args used
     */
    public static void main(String[] args) {
        System.out.println("--- SYSTEM PROPERTIES ---");
        final Enumeration spe = System.getProperties().propertyNames();
        while (spe.hasMoreElements()) {
            final String key = (String) spe.nextElement();
            final String value = System.getProperty(key);
            System.out.println("property:" + key + "=\"" + value + "\"");
        }
        System.out.println("--- ENVIRONMENT ---");
        for (Map.Entry<String, String> entry : System.getenv().entrySet()) {
            System.out.println("env:" + entry.getKey() + "=\"" + entry.getValue() + "\"");
        }
    }
}
