package net.sf.buildbox.util;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;

/**
 * Task that executes a process.
 * The added value compared to {@link ProcessBuilder} is that the caller either does not have to care about processing stdout and stderr streams,
 * or can  use line-oriented {@link #stdout(String)} and {@link #stderr(String)} methods to handle it.
 * <p>You can use system property "<code>net.sf.buildbox.util.CommandLineExec.prepend</code>" to specify arguments to prepend in order to invoke executables.
 * If unspecified, on Windows, arguments "cmd /C" are automatically prepended. Future versions may come with further variants.
 * </p>
 */
public class CommandLineExec<T> implements Callable<Integer> {
    private static final Collection<String> shellPrepend = new ArrayList<String>();
    static {
        final String prepend = System.getProperty("net.sf.buildbox.util.CommandLineExec.prepend");
        if (prepend != null) {
            for (StringTokenizer tok = new StringTokenizer(prepend, " "); tok.hasMoreTokens();) {
                final String arg = tok.nextToken();
                shellPrepend.add(arg);
            }
        } else {
            // OS-specific tweaks
            final String osName = System.getProperty("os.name").toLowerCase();
            if (osName.startsWith("win")) {
                shellPrepend.add("cmd");
                shellPrepend.add("/C");
            }
        }
    }
    private File basedir;
    private final String executable;
    private List<String> argList = new ArrayList<String>();
    private boolean failOnError = true;

    /**
     * Allows adding some extra arguments.
     * @param args -
     * @see #addArguments(List)
     */
    public final void addArguments(String... args) {
        argList.addAll(Arrays.asList(args));
    }

    /**
     * Allows adding some extra arguments.
     * @param args -
     * @see #addArguments(String[])
     */
    public final void addArguments(List<String> args) {
        argList.addAll(args);
    }

    /**
     * Sets the directory where the process will run.
     * If set, {@link #call()} will {@link #log} it as "cd ((basedir))".
     * @param basedir see {@link ProcessBuilder#directory(File)}  
     */
    public final void setBasedir(File basedir) {
        this.basedir = basedir;
    }

    /**
     * Controls exitCode evaluation behaviour.
     * By default, exitCode is only logged
     * @param failOnError  if true, {@link #call()} fails with {@link java.io.IOException} if returned process exitCode is not 0.
     */
    public final void setFailOnError(boolean failOnError) {
        this.failOnError = failOnError;
    }

    public CommandLineExec(String executable, String... args) {
        this.executable = executable;
        argList.addAll(Arrays.asList(args));
    }

    /**
     * Executes the prepared shell command.
     * Passes standard output to {@link #stdout(String)}, standard error to {@link #stderr(String)} and other stuff to {@link #log(String)}.
     *
     * @see #setBasedir(java.io.File)
     * @see #setFailOnError(boolean)
     * @see #addArguments(java.util.List)
     * @see #addArguments(String[])
     *
     * @return process exit code. If {@link #setFailOnError(boolean) failOnError} is set, throws {@link ExecIOException} instead.
     * @throws InterruptedException -
     * @throws IOException  see {@link ProcessBuilder#start()}
     * @throws ExecIOException  if exitCode differs from 0 and {@link #setFailOnError(boolean) failOnError} is set
     */
    public final Integer call() throws InterruptedException, IOException {
        try {
            final ThreadLocal<Integer> exitCodeHolder = new ThreadLocal<Integer>();
            invoke(exitCodeHolder);
            return exitCodeHolder.get();
        } catch (InterruptedException e) {
            throw e;
        } catch (IOException e) {
            throw e;
        } catch (Exception e) {
            throw new IOException(e.getMessage());
        }
    }

    public final T invoke() throws Exception {
        return invoke(null);
    }

    private T invoke(ThreadLocal<Integer> exitCodeHolder) throws Exception {
        final List<String> cmdline = new ArrayList<String>();
        cmdline.addAll(shellPrepend);
        cmdline.add(executable);
        cmdline.addAll(argList);

        final ProcessBuilder pb = new ProcessBuilder();
        if (basedir != null) {
            pb.directory(basedir);
            log("cd " + basedir);
        }
        log(toString(cmdline, " "));
        pb.command(cmdline);
        overrideEnv(pb.environment());
        final Process proc = pb.start();
        try {
            // start output consumers
            final ExecutorService es = Executors.newFixedThreadPool(2);
            final Future<T> stdoutFuture = es.submit(createStdOutConsumer(proc.getInputStream()));
            es.submit(createStdErrConsumer(proc.getErrorStream()));
            // start input feeder
            stdin(proc.getOutputStream());
            // wait for the process
            proc.waitFor();
            // stop consumers
            es.shutdownNow();
            final int exitCode = proc.exitValue();
            if ((exitCode != 0)) {
                if (exitCodeHolder != null) {
                    exitCodeHolder.set(exitCode);
                }
                if (failOnError) {
                    throw new ExecIOException(executable, exitCode);
                }
                log("exitCode = " + exitCode);
                return null;
            }
            return stdoutFuture.get();
        } catch (ExecutionException e) {
            try {
                throw e.getCause();
            } catch (Exception ee) {
                throw ee;
            } catch (Throwable ee) {
                throw e;
            }
        } finally {
            proc.destroy();
        }
    }

    protected Callable<T> createStdOutConsumer(InputStream is) {
        return new StreamConsumer(is, false);
    }

    protected Callable<T> createStdErrConsumer(InputStream is) {
        return new StreamConsumer(is, true);
    }

    protected void stdin(OutputStream outputStream) throws IOException {
        // override to feed the process' stdin with data
    }

    /**
     * Exposes process environment for modifications.
     * By default, no modifications are done - env is just printed using {@link #logEnv(java.util.Map)}.
     * @param env  the environment map that can be effectively altered; see {@link ProcessBuilder#environment()}.
     */
    protected void overrideEnv(Map<String, String> env) {
        logEnv(env);
    }

    /**
     * Prints the environment, one by one, using {@link #log} method.
     * Each call prints one key/value pair in format "{key}={value}".
     * @param env -
     */
    protected void logEnv(Map<String, String> env) {
//        defaultLogEnv(env);
    }

    protected void defaultLogEnv(Map<String, String> env) {
        final List<String> sortedKeys = new ArrayList<String>(env.keySet());
        Collections.sort(sortedKeys);
        for (String key : sortedKeys) {
            log(key + "=" + env.get(key));
        }
    }

    /**
     * Prints an extra information before or after process execution.
     * <p>Override to use different logging device than the default {@link System#err}.</p>
     *
     * @param message - the message to log, usually a single line
     */
    protected void log(String message) {
        System.err.println("INFO: " + message);
    }

    /**
     * Used to log exceptions from internal stream consumers.
     * <p>Override to use different logging device than the default {@link System#err}.</p>
     * @param throwable -
     */
    protected void printStackTrace(Throwable throwable) {
        throwable.printStackTrace();
    }

    /**
     * Used to log one line output from stdout stream.
     * <p>Override to use different logging device than the default {@link System#out}.</p>
     * @param line -
     */
    protected void stdout(String line) {
        System.out.println(line);
    }

    /**
     * Used to log one line output from stderr stream.
     * <p>Override to use different logging device than the default {@link System#err}.</p>
     * @param line -
     */
    protected void stderr(String line) {
        System.err.println(line);
    }

    /**
     * Merges string representation of collection items into one string, using provided delimiter.
     * For example, used for presenting commandline arguments, where comma delimiter would be annoying.
     * @param coll -
     * @param delimiter -
     * @return string of merged items using the desired delimiter
     */
    public static String toString(Collection<?> coll, String delimiter) {
        final StringBuilder sb = new StringBuilder();
        for (Object item : coll) {
            if (sb.length() > 0) {
                sb.append(delimiter);
            }
            sb.append(item);
        }
        return sb.toString();
    }

    @Override
    public String toString() {
        return "CommandLineExec{" + executable + " " + toString(argList, " ") + "}";
    }

    private class StreamConsumer implements Callable<T> {
        private final BufferedReader bufferedReader;
        private final boolean isStdErr;

        public StreamConsumer(InputStream is, boolean stdErr) {
            this.isStdErr = stdErr;
            this.bufferedReader = new BufferedReader(new InputStreamReader(is));
        }

        public T call() throws IOException {
                String line = bufferedReader.readLine();
                while (line != null) {
                    if (isStdErr) {
                        stderr(line);
                    } else {
                        stdout(line);
                    }
                    line = bufferedReader.readLine();
                }
            return null;
        }

        @Override
        public String toString() {
            return isStdErr ? "ERR" : "OUT";
        }
    }


    public static class ExecIOException extends IOException {
        private final String executable;
        private final int exitCode;

        private ExecIOException(String executable, int exitCode) {
            super(String.format("Execution of %s failed with exitCode %d", executable, exitCode));
            this.executable = executable;
            this.exitCode = exitCode;
        }

        public final String getExecutable() {
            return executable;
        }

        public final int getExitCode() {
            return exitCode;
        }
    }
}
