package net.sf.buildbox.util;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.*;

public final class DateUtils {

    private DateUtils() {
        // not instantiable
    }

    private static final ThreadLocal<Map<String, DateFormat>> userDateFormats = new ThreadLocal<Map<String, DateFormat>>();

    /**
     * Parse ISO-8601 (and XML-Schema) compliant date string.
     *
     * @param date String containing the date representation
     * @return java.util.Date parsed date
     * @throws java.text.ParseException if parameter doesn't represent valid ISO compliant date
     * @todo replace with better styled and more ISO compliant method (zones, etc..)
     */
    public static Date parseIsoDate(String date) throws ParseException {
        return IsoFormat.parseIsoDate(date);
    }

    /**
     * Format date into ISO-8601 (~ XML-Schema) representation.
     *
     * @param date      date to format
     * @param isoFormat specifies format
     * @return String containing the ISO representation
     */
    public static String formatIsoDate(Date date, IsoFormat isoFormat) {
        return (date != null) ? isoFormat.format(date) : null;
    }

    /**
     * Format date into ISO-8601 (and XML-Schema) representation.
     * Uses <code>IsoFormat.DEFAULT</code> format.
     *
     * @param date date to format
     * @return String containing the ISO representation
     * @see #formatIsoDate(java.util.Date,IsoFormat)
     */
    public static String formatIsoDate(Date date) {
        return formatIsoDate(date, IsoFormat.DEFAULT);
    }

    /**
     * Formats date using specific date format (cached, thread safe)
     *
     * @param date       date to format
     * @param dateFormat format to use for formatting the date (null for default Iso format)
     * @return formatted date
     */
    public static String formatDate(Date date, String dateFormat) {

        Map<String, DateFormat> dateFormats = userDateFormats.get();
        if (dateFormats == null) {
            dateFormats = new HashMap<String, DateFormat>(3);
            userDateFormats.set(dateFormats);
        }
        if (dateFormat == null)
            dateFormat = "";
        DateFormat sdf = dateFormats.get(dateFormat);
        if (sdf == null) {
            sdf = "".equals(dateFormat) ? IsoFormat.DEFAULT.getDateFormat() : new SimpleDateFormat(dateFormat);
            dateFormats.put(dateFormat, sdf);
        }
        return sdf.format(date);
    }

    /**
     * Parses date using specific date format (cached, thread safe)
     *
     * @param date       date to parse
     * @param dateFormat format to use for parsing the date (null for default Iso format)
     * @return parsed date
     * @throws java.text.ParseException when the string does not match specified format
     */
    public static Date parseDate(String date, String dateFormat) throws ParseException {

        Map<String, DateFormat> dateFormats = userDateFormats.get();
        if (dateFormats == null) {
            dateFormats = new HashMap<String, DateFormat>(3);
            userDateFormats.set(dateFormats);
        }
        if (dateFormat == null)
            dateFormat = "";
        DateFormat sdf = dateFormats.get(dateFormat);
        if (sdf == null) {
            sdf = "".equals(dateFormat) ? IsoFormat.DEFAULT.getDateFormat() : new SimpleDateFormat(dateFormat);
            dateFormats.put(dateFormat, sdf);
        }
        return sdf.parse(date);
    }

    /**
     * Specifies date format.
     */
    public static final class IsoFormat {

        private static List<IsoFormat> formats = new ArrayList<IsoFormat>();

        private ThreadLocal dateFormat;

        private IsoFormat(final String dateFormat) {
            this.dateFormat = new ThreadLocal() {
                protected Object initialValue() {
                    DateFormat format = new SimpleDateFormat(dateFormat);
                    format.setLenient(false);
                    return format;
                }
            };
            formats.add(this);
        }

        private DateFormat getDateFormat() {
            return (DateFormat) dateFormat.get();
        }

        private String format(Date date) {
            return getDateFormat().format(date);
        }

        private static Date parseIsoDate(String date) throws ParseException {
            if (date == null) {
                throw new ParseException("Parsed date can't be null", 0);
            }
            date = date.trim();

            if (date.length() >= 1 && date.charAt(date.length() - 1) == 'Z') {
                date = date.substring(0, date.length() - 1) + "GMT+00:00";
            } else if (date.length() >= 6
                    && (date.charAt(date.length() - 6) == '+' || date.charAt(date.length() - 6) == '-')
                    && date.charAt(date.length() - 3) == ':') {
                date = new StringBuffer(date).insert(date.length() - 6, "GMT").toString();
            }

            for (IsoFormat format : formats) {
                ParsePosition pos = new ParsePosition(0);
                Date d = format.getDateFormat().parse(date, pos);
                if (d != null && pos.getIndex() == date.length()) {
                    return d;
                }
            }
            throw new ParseException("No date format available to parse this: '" + date + "'", -1);
        }

        /**
         * Represents yyyy-MM-dd'T'HH:mm:ss.SSS format.
         */
        public static final IsoFormat WITH_MILLIS = new IsoFormat("yyyy-MM-dd'T'HH:mm:ss.SSS");
//        private static final IsoFormat WITH_MILLIS_ZONE = new IsoFormat("yyyy-MM-dd'T'HH:mm:ss.SSSz");

        /**
         * Represents yyyy-MM-dd'T'HH:mm:ss format.
         */
        public static final IsoFormat DEFAULT = new IsoFormat("yyyy-MM-dd'T'HH:mm:ss");
//        private static final IsoFormat DEFAULT_ZONE = new IsoFormat("yyyy-MM-dd'T'HH:mm:ssz");

//        private static final IsoFormat WITHOUT_SECONDS = new IsoFormat("yyyy-MM-dd'T'HH:mm");

        /**
         * Represents yyyy-MM-dd format.
         */
        public static final IsoFormat DATE_ONLY = new IsoFormat("yyyy-MM-dd");

    }
}
