package net.sf.buildbox.util;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Collection;
import java.util.regex.Pattern;

/**
 * Utility class with methods for accessing trivial file formats.
 */
public final class SimpleTextFile {
    private static String NEWLINE = System.getProperty("line.separator");

    /**
     * Designed for use with {@link #readAsLines(java.io.File, java.util.regex.Pattern)}.
     * Removes all empty lines and lines starting with hash.
     */
    public static final Pattern BASH_LIKE_LINES = Pattern.compile("^(\\s*|#.*)$");

    private SimpleTextFile() {
        // not instantiable
    }

    /**
     * Reads lines from a file as a collection of strings.
     *
     * @param inputFile -
     * @param excludeLinePattern pattern matching lines that should be skipped; see {@link #BASH_LIKE_LINES}. If null, no skipping occurs.
     * @return List of strings
     * @throws IOException -
     */
    public static List<String> readAsLines(File inputFile, Pattern excludeLinePattern) throws IOException {
        BufferedReader r = null;
        final List<String> lines = new ArrayList<String>();
        try {
            r = new BufferedReader(new FileReader(inputFile));
            String line = r.readLine();
            while (line != null) {
                if (excludeLinePattern == null || !excludeLinePattern.matcher(line).matches()) {
                    lines.add(line);
                }
                line = r.readLine();
            }
        } finally {
            if (r != null) {
                r.close();
            }
        }
        return lines;
    }

    public static void writeAsLines(File outputFile, Collection<String> lines) throws IOException {
        final FileWriter w = new FileWriter(outputFile);
        try {
            for (String line : lines) {
                w.write(line + NEWLINE);
            }
        } finally {
            w.close();
        }
    }

    public static void simpleSave(File file, String text) throws IOException {
        final BufferedWriter w = new BufferedWriter(new FileWriter(file));
        try {
            w.write(text);
        } finally {
            w.close();
        }
    }

}
