package net.sf.buildbox.util;

import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;

/**
 * Executes a task in separate thread and waits for its completion, providing keepalive logging opportunity in the meantime.
 * Unlike {@link net.sf.buildbox.util.SlowTaskMon}, this is designed for use in managed threads that are required to indicate its life by keep-alive logging.
 * @param <T>
 */
public abstract class SlowTask<T> implements Callable<T> {
    private final Callable<T> callable;
    protected final String actionName;
    private final long firstPeriod;
    private final long nextPeriod;
    private volatile boolean stopPlease;

    public SlowTask(String actionName, long period, Callable<T> callable) {
        this(actionName, period, period, callable);
    }

    public SlowTask(String actionName, long firstPeriod, long nextPeriod, Callable<T> callable) {
        this.callable = callable;
        this.actionName = actionName;
        this.firstPeriod = firstPeriod;
        this.nextPeriod = nextPeriod;
    }

    public synchronized final T call() throws Exception {
        final FutureTask<T> ft = new FutureTask<T>(callable) {
            @Override
            protected void done() {
                stopPlease = true;
                synchronized (callable) {
                    callable.notifyAll();
                }
            }
        };
        final Thread runThread = new Thread(ft, actionName + ".run");
        //
        try {
            onStarting();
            runThread.start();
            synchronized (callable) {
                stopPlease = false;
                Thread.sleep(firstPeriod);
                while (!stopPlease) {
                    stillWorking();
                    callable.wait(nextPeriod);
                }
            }
            return ft.get();
        } finally {
            runThread.join();
            onCompleted();
        }
        //
    }

    protected void onStarting() {
        //
    }

    public abstract void stillWorking();

    protected void onCompleted() {
        //
    }

}
