package net.sf.buildbox.util;

import java.io.File;
import java.io.IOException;
import java.io.FilenameFilter;
import java.io.FileFilter;
import java.util.Set;
import java.util.LinkedHashSet;

/**
 * Access a file using file-locking mechanism.
 * The rule is simple: if there is a lockfile, somebody is writing to it and we have to wait.
 * While waiting, we watch the size and modification time, and keep waiting as long as it keeps changing.
 * When changes cease for "longer" period, we give up. By default that means throwing
 *  an exception.
 * @deprecated use {@link net.sf.buildbox.util.ExtFileLock} instead. 
 */
@Deprecated
public abstract class SynchronizedFile {
    private final File file;
    private final File lockdir;
    private final long quantum;
    private final int noChangeTolerance;

    public SynchronizedFile(File file) {
        this(file, 250, 20);
    }
    public SynchronizedFile(File file, long quantum, int noChangeTolerance) {
        this.file = file;
        this.quantum = quantum;
        this.noChangeTolerance = noChangeTolerance;
        lockdir = new File(file.getAbsolutePath() + ".lock");
    }

    public final boolean safeRead() throws InterruptedException, IOException {
        final File lockfile = lock(false);
        if (lockfile == null) return false;
        try {
            readingBlock();
        } finally {
            unlock(lockfile, false);
        }
        return true;
    }

    public final boolean safeWrite() throws InterruptedException, IOException {
        final long lastModified = file.lastModified();
        final File lockfile = lock(true);
        if (lockfile == null) return false;
        try {
            if (file.lastModified() != lastModified) {
                throw new IOException(file + ": file was modified while waiting for write lock, refusing to overwrite");
            }
            writingBlock();
        } finally {
            unlock(lockfile, true);
        }
        return true;
    }

    private File lock(boolean forWriting) throws InterruptedException, IOException {
        while (true) {
            // if we succeeded to create directory, we are the lucky lockers
            if (lockdir.mkdir()) break;
            if (!forWriting) {
                // for reading, if there is one other reader, we can also lock it for reading
                // otherwise it's not sure that it is reading lock so we prefer to wait
                if (otherReaderCount() > 0) break;
            }
            // failed to create lockdir, let's wait
            int noChangeDetectedCnt = 0;
            while (lockdir.exists()) {
                final long lastModified = file.lastModified();
                final long length = file.length();
                Thread.sleep(quantum);
                if (lastModified == file.lastModified() && length == file.length()) {
                    noChangeDetectedCnt++;
                } else {
                    noChangeDetectedCnt = 0;
                }

                if (noChangeDetectedCnt > noChangeTolerance) {
                    giveUp();
                    // failed to lock but not throwing exception
                    return null;
                }
            }
        }
        // file is unlocked for the desired purpose, and locked for me - let's indicate purpose
        final String threadName = Thread.currentThread().getName();
        final File lockfile = File.createTempFile(forWriting ? ".W." : "rdr.", "." + threadName.replaceAll("[:/\\\\ ]", "_"), lockdir);
        SimpleTextFile.simpleSave(lockfile, String.format("%tFT%1$tT.%1$tL Thread: %s", System.currentTimeMillis(), threadName));
        return lockfile;
    }

    private int otherReaderCount() {
        if (!lockdir.isDirectory()) {
            return 0;
        }
        final String[] readerLocks = lockdir.list(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return !name.startsWith(".");
            }
        });
        return readerLocks == null ? 0 : readerLocks.length;
    }

    private void unlock(File lockfile, boolean afterWriting) throws IOException {
        lockfile.delete();
        if (otherReaderCount() == 0) {
            lockdir.delete();
            if (lockdir.exists()) {
                final Set<File> remainingFiles = new LinkedHashSet<File>();
                lockdir.listFiles(new FileFilter() {
                    public boolean accept(File file) {
                        final String fn = file.getName();
                        if (fn.equals(".")) return false;
                        if (fn.equals("..")) return false;
                        remainingFiles.add(file);
                        return false;
                    }
                });
                unlockFailed(lockfile, afterWriting, remainingFiles);
            }
        }
    }

    protected void unlockFailed(File lockfile, boolean afterWriting, Set<File> files) throws IOException {
        for (File file : files) {
            System.err.println(String.format("%tFT%1$tT.%1$tL ERROR unlock prevented by: %s", file.lastModified(), file.getAbsolutePath()));
        }
        throw new IOException(lockfile + ": failed to unlock " + (afterWriting ? "after writing" : "last reader"));
    }

    /**
     * Override this with your code reading the file.
     *
     * @throws java.io.IOException -
     */
    protected void readingBlock() throws IOException {
        throw new IOException(file + ": readingBlock not implemented");
    }

    /**
     * Override this with your code modifying the file.
     *
     * @throws java.io.IOException -
     */
    protected void writingBlock() throws IOException {
        throw new IOException(file + ": writingBlock not implemented");
    }

    protected void giveUp() throws IOException {
        throw new IOException(file + ": still locked by " + lockdir + ", giving up");
    }
}
